# <@LICENSE>
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to you under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </@LICENSE>
#

#
# Version: 20091122
# This is plugin/module is Rutgers Hackery(tm)
# Created by Eric Rivas <kc2hmv@nbcs.rutgers.edu>
# Meta tag checking and rulang_utf8 added by Jarek Sedlacek <jarek@nbcs.rutgers.edu>

#
# SpamAssassin Conf, local.cf:
#  loadplugin RULang /path/to/RULang.pm
#  header RU_LANG_HEADER eval:rulang_header()
#  header RU_LANG_UTF8	 eval:rulang_utf8()
#  score RU_LANG_HEADER 0.01
#  score RU_LANG_UTF8 0.01
#  add_header all RU-Lang _RULANG_


package RULang;
#for utf8 support in perl < 5.8
use Unicode::UTF8simple;

use Mail::SpamAssassin::Plugin;

use strict;
use warnings;
# We inherit from Plugin.
our @ISA;
@ISA = qw(Mail::SpamAssassin::Plugin);

# Lang-Charset hash
# langtype => charsets
our %charsets = (
    'ru' => "KOI8-R WINDOWS-1251 ISO-8858-5 KOI8-U",
    'cn' => "GB2312 HZ-GB-2312 BIG5 GB18030 GBK"
);

#'bad' utf8 chars
our %utf8chars = (
	#note: that is NOT the normal 'a' 'e' etc on the line below.
	#They are russian chars that just happen to look similar
	#utf8 Russian alphabet (source: http://www.slovo.info/testuni.htm)
	'ru' => "а б в г д е ё ж з и й к л м н о п р с т у ф х ц ч ш ч ь ы ъ э ю я А Б В Г 
Д Е Ё Ж З И Й К Л М Н О П Р С Т У Ф Х Ц Ч Ш Щ Ь Ы Ъ Э Ю Я",
	#utf8 chinese chars
	'cn' => "诶 比 西 迪 伊 艾 弗 吉 尺 杰 开 勒 马 娜 哦 屁 吾 儿 丝 提 维 豆 贝 尔 克 斯 贼 德 你 好！公 司 受 理 各 好！公司 业 单 位 税＊票 价 格 优 惠 验 证 后 付 电 话 张 "
); 
sub new {
    my ($class, $mailsa) = @_;
    my $self;

    $class = ref($class) || $class;
    $self = $class->SUPER::new($mailsa);
    bless($self, $class);

    $self->register_eval_rule("rulang_header");
    $self->register_eval_rule("rulang_utf8");

    return $self;
}
# get percentage of the message that consists of utf8 chars in bad_chars
# if its over a certain percentage, mark it as spam
# punctuation and whitespace is excluded
# this also fires when utf8 is detected in the subject line
sub rulang_utf8(){
	#maximum percent of the body that can be utf8 russian or chinese chars
	#and more than this, and its flagged as spam
	my $max_percent = 60;
	
	Mail::SpamAssassin::Plugin::dbg("RULang: Begin rulang_utf8");
	
	my ($self, $permsgstatus) = @_;
	my $arraybody = $permsgstatus->get_decoded_body_text_array();
	my $strbody = join(" ",@$arraybody);
	my $subject = $permsgstatus->get("subject");
	
	#trim all spaces
	$strbody =~ s/\s+//g;
	$subject =~ s/\s+//g;
	#remove punctuation
	$strbody =~ tr/.!?;:(),{}//d;
	$subject =~ tr/.!?;:(),{}//d;
	
	my $isspam = 0;
	my $utf8_char_count = 0;
	
	#get number of times each utf8 character appears
	for my $key (keys(%utf8chars)) {
		use bytes; #we use bytes here for multibyte chars
		my @targetChar = split(/ /,$utf8chars{$key});
		for my $target (@targetChar){
			my $num_occurences = 0;
			# because we're using bytes, not utf8, we have to get the
			# number of characters manually, instead of using tr() or the like
			while ($strbody =~ /$target/g) { $num_occurences++;}
			if ($subject =~ /$target/g) {#if the subject matches any of the bad chars, fire RULang_utf8
				$isspam=1;
				Mail::SpamAssassin::Plugin::dbg("RULang: utf8 detected in the subject line");
			}
			$utf8_char_count += $num_occurences;
		}
    }
    
    	#we have to decode the string to get the length correctly
	my $uref = new Unicode::UTF8simple;
	$strbody=$uref->fromUTF8("iso-8859-1",$strbody);
	$strbody =~ s/_//g;
	my $bodylength = length( $strbody);
	# the utf8 chars got counted twice, because they have two bytes
	$bodylength -= $utf8_char_count;
	my $percent = ($utf8_char_count / $bodylength) * 100;
	Mail::SpamAssassin::Plugin::dbg("RULang: $utf8_char_count bad utf8 chars in the body");
	Mail::SpamAssassin::Plugin::dbg("RULang: $bodylength total chars in the body");
	Mail::SpamAssassin::Plugin::dbg("RULang: body is $percent% bad utf8 chars");
    
    if ($percent >= $max_percent){
		$isspam = 1;
	}

	Mail::SpamAssassin::Plugin::dbg("RULang: End rulang_utf8");
	return $isspam;
}

sub rulang_header {
    my ($self, $permsgstatus) = @_;
    my @cslocales;
    my $isspam = 0;

    Mail::SpamAssassin::Plugin::dbg("RULang: Begin rulang_header");

    # Search the headers
    for my $h (qw(From Subject)) {
        my $lang;
        my $hdr = $permsgstatus->get("$h:raw");
        Mail::SpamAssassin::Plugin::dbg("RULang: Checking  $h:$hdr");
        while ($hdr =~ /=\?(.+?)\?.\?.*?\?=/g) {
            Mail::SpamAssassin::Plugin::dbg("RULang: $h charset: $1");
            $lang = get_all_lang_from_cs($1);
            if ($lang ne "") {
                push(@cslocales, $lang);
            }
        }
    }

    # Content-Type is special
    {
        my $lang;
        my $cs;
        my $hdr = $permsgstatus->get("Content-Type:raw", "");
            unless ($hdr eq "") {
                $hdr =~ s/\n/ /g;
            	Mail::SpamAssassin::Plugin::dbg("RULang: Content-Type: $hdr");
            	$cs = get_charset_from_ct_line($hdr);
            	Mail::SpamAssassin::Plugin::dbg("RULang: CS detected from CT: $cs");
            	if ($cs ne "") {
                	$lang = get_all_lang_from_cs($cs);
                	if ($lang ne "") {
                    		push(@cslocales, $lang);
                	}
		}
	    
	}
     }
    # Content-Description is also special
    {
        my $lang;
        my $cs;
        my $hdr = $permsgstatus->get("Content-Description:raw", "");
            unless ($hdr eq "") {
                $hdr =~ s/\n/ /g;
            	Mail::SpamAssassin::Plugin::dbg("RULang: Content-Description: $hdr");
            	$cs = get_charset_from_cd_line($hdr);
            	Mail::SpamAssassin::Plugin::dbg("RULang: CS detected from CD: $cs");
            	if ($cs ne "") {
                	$lang = get_all_lang_from_cs($cs);
                	if ($lang ne "") {
                    		push(@cslocales, $lang);
                	}
		}
	    }
    }
    
    #Check <meta> tags
	Mail::SpamAssassin::Plugin::dbg("RULang: Checking meta tags");
	my $arraybody = $permsgstatus->get_decoded_body_text_array();
	my $strbody = join(" ",@$arraybody);
	
	#if any of the charsets are found within meta tags in the body
	#push them on the stack
	for my $key (keys(%charsets)) {
		my @targetCharSets = split(/ /,$charsets{$key});
		for my $target (@targetCharSets){
			if ($strbody =~ /<meta.*?$target.*?>/gi) {
				push(@cslocales, $target);
			}
		}
    }
    Mail::SpamAssassin::Plugin::dbg("RULang: End meta tag check");
    

    # Set template tag
    Mail::SpamAssassin::Plugin::dbg("RULang: Header cslocales: @cslocales");
    if (scalar(@cslocales) != 0) {
        my $sets = join(' ', uniq_list(@cslocales));
        $permsgstatus->set_tag("RULANG", $sets);
        $isspam = 1;
    } else {
        $permsgstatus->set_tag("RULANG", "");
        $isspam = 0;
    }	
	
    Mail::SpamAssassin::Plugin::dbg("RULang: End rulang_header");
    return $isspam;
}

sub get_all_lang_from_cs {
    my $cs = shift || "";
    my @langs;

    for my $key (keys(%charsets)) {
        if ($charsets{$key} =~ /\b$cs\b/i) {
            push(@langs, $key);
        }
    }

    if (scalar(@langs) > 0) {
        return join(' ', @langs);
    } else {
        return "";
    }
}

sub get_charset_from_cd_line{
	my $cd_line = $_[0];
	for my $key (keys(%charsets)) {
		my @targetCharSets = split(/ /,$charsets{$key});
		for my $target (@targetCharSets){
			if ($cd_line =~ /$target/gi) {
					
				return $target;	
			}
		}
	}	

}

sub get_charset_from_ct_line {
    my $type = shift;
    if ($type =~ /charset="([^"]+)"/i) { return $1; }
    if ($type =~ /charset='([^']+)'/i) { return $1; }
    if ($type =~ /charset=(\S+)/i) { return $1; }
    return "";
}


sub uniq_list {
    my @cluter = shift;
    my %seen;

    return sort(grep !$seen{$_}++, @cluter);
}

1;

